from typing import List
import numba


@numba.jit(nopython=True)
def shingling_k(s: str, k: int) -> List[List[str]]:
    """Convert a string to a list of k sequences with k-shingles

    Parameters:
    -----------
    s: str
        The raw string

    k: int
        The parameter must be k>=1

    Returns:
    --------
    List[List[str]]
        A list of k sequences. The 1st sublist contains the chars of the
          string s, e.g. ['a', 'b', 'c'] for s="abc". The k=2 list containd
          2-shingles, e.g. ['ab', 'bc'].

    Example:
    --------
        import kshingle as ks
        shingles = ks.shingling_k("abc", k=2)
    """
    shingles = []
    q = len(s)
    for n in range(1, k + 1):
        shingles.append([s[i:(i + n)] for i in range(q - n + 1)])
    return shingles


@numba.jit(nopython=True)
def shingling_range(s: str, n_min: int, n_max: int) -> List[List[str]]:
    """Convert a string to a list of k sequences with k-shingles

    Parameters:
    -----------
    s: str
        The raw string

    n_min: int
        The lower bound k range [n_min, n_max]

    n_max: int
        The upper bound of the k range [n_min, n_max]

    Returns:
    --------
    List[List[str]]
        A list of k sequences. The k=1 sublist contains the chars of the
          string s, e.g. ['a', 'b', 'c', 'd'] for s="abcd". The k=3 list
          contains 3-shingles, e.g. ['abc', 'bcd'].

    Example:
    --------
        import kshingle as ks
        shingles = ks.shingling_range("abcd", [2, 3])
    """
    # correct wrong inputs
    n_max_ = max(0, n_max)
    n_min_ = max(1, n_min)
    # start to loop
    shingles = []
    q = len(s)
    for n in range(n_min_, n_max_ + 1):
        shingles.append([s[i:(i + n)] for i in range(q - n + 1)])
    return shingles


def shingling_list(s: str, klist: List[int]) -> List[List[str]]:
    """Convert a string to a list of k sequences with k-shingles

    Parameters:
    -----------
    s: str
        The raw string

    klist: List[int]
        A list of specific k values, e.g. klist=[1, 3, 6]

    Returns:
    --------
    List[List[str]]
        A list of k sequences. The k=1 sublist contains the chars of the
          string s, e.g. ['a', 'b', 'c', 'd'] for s="abcd". The k=3 list
          contains 3-shingles, e.g. ['abc', 'bcd'].

    Example:
    --------
        import kshingle as ks
        shingles = ks.shingling_list("abcd", klist=[1, 3])
    """
    return shingling_list_numba(s, numba.typed.List(klist))


@numba.jit(nopython=True)
def shingling_list_numba(s: str,
                         klist: numba.typed.List
                         ) -> List[List[str]]:
    shingles = []
    q = len(s)
    for n in klist:
        if n > 0:
            shingles.append([s[i:(i + n)] for i in range(q - n + 1)])
    return shingles



__version__ = '0.6.5'

from .shingling import (shingling_k, shingling_list, shingling_range)
from .shingleset import (shingleset_k, shingleset_range, shingleset_list)
from .vocab import (
    identify_vocab, upsert_word_to_vocab, encoded_with_vocab,
    shrink_k_backwards)
from .wildcard import wildcard_shinglesets
from .metrics import jaccard, jaccard_strings



import kshingle as ks


def test1():
    shingles = ks.shingling_k("", 1)
    assert shingles == [[]]

    shingles = ks.shingling_k(" ", 1)
    assert shingles == [[" "]]

    shingles = ks.shingling_k(" ", 2)
    assert shingles == [[" "], []]


def test2():
    shingles = ks.shingling_k(" ", 0)
    assert shingles == []

    shingles = ks.shingling_k(" ", -1)
    assert shingles == []


def test3():
    shingles = ks.shingling_k("12345", 0)
    assert shingles == []

    shingles = ks.shingling_k("12345", 1)
    assert shingles == [['1', '2', '3', '4', '5']]

    shingles = ks.shingling_k("12345", 2)
    assert shingles == [
        ['1', '2', '3', '4', '5'],
        ['12', '23', '34', '45']]

    shingles = ks.shingling_k("12345", 3)
    assert shingles == [
        ['1', '2', '3', '4', '5'],
        ['12', '23', '34', '45'],
        ['123', '234', '345']]

    shingles = ks.shingling_k("12345", 4)
    assert shingles == [
        ['1', '2', '3', '4', '5'],
        ['12', '23', '34', '45'],
        ['123', '234', '345'],
        ['1234', '2345']]

    shingles = ks.shingling_k("12345", 5)
    assert shingles == [
        ['1', '2', '3', '4', '5'],
        ['12', '23', '34', '45'],
        ['123', '234', '345'],
        ['1234', '2345'],
        ['12345']]

    shingles = ks.shingling_k("12345", 6)
    assert shingles == [
        ['1', '2', '3', '4', '5'],
        ['12', '23', '34', '45'],
        ['123', '234', '345'],
        ['1234', '2345'],
        ['12345'], []]


def test11():
    shingles = ks.shingling_range("", 0, 1)
    assert shingles == [[]]

    shingles = ks.shingling_range(" ", 0, 1)
    assert shingles == [[" "]]

    shingles = ks.shingling_range(" ", 0, 2)
    assert shingles == [[" "], []]

    shingles = ks.shingling_range("", 1, 1)
    assert shingles == [[]]

    shingles = ks.shingling_range(" ", 1, 1)
    assert shingles == [[" "]]

    shingles = ks.shingling_range(" ", 1, 2)
    assert shingles == [[" "], []]


def test12():
    shingles = ks.shingling_range("", -10, -10)
    assert shingles == []

    shingles = ks.shingling_range("", 0, 0)
    assert shingles == []

    shingles = ks.shingling_range("", 4, 1)
    assert shingles == []


def test31():
    shingles = ks.shingling_list("", [1])
    assert shingles == [[]]

    shingles = ks.shingling_list(" ", [1])
    assert shingles == [[" "]]

    shingles = ks.shingling_list(" ", [1, 2])
    assert shingles == [[" "], []]


def test32():
    shingles = ks.shingling_list(" ", [0])
    assert shingles == []

    shingles = ks.shingling_list(" ", [-1])
    assert shingles == []


def test33():
    shingles = ks.shingling_list("12345", [5, 6])
    assert shingles == [['12345'], []]

    shingles = ks.shingling_list("12345", [0, 3, 6])
    assert shingles == [['123', '234', '345'], []]  # this is a problem!



from typing import List


def shingleset_k(s: str, k: int) -> set:
    shingles = []
    q = len(s)
    for n in range(1, k + 1):
        shingles.append(set([s[i:(i + n)] for i in range(q - n + 1)]))
    return set.union(*shingles)


def shingleset_range(s: str, n_min: int, n_max: int) -> set:
    # correct wrong inputs
    n_max_ = max(0, n_max)
    n_min_ = max(1, n_min)
    # start to loop
    shingles = []
    q = len(s)
    for n in range(n_min_, n_max_ + 1):
        shingles.append(set([s[i:(i + n)] for i in range(q - n + 1)]))
    return set.union(*shingles)


def shingleset_list(s: str, klist: List[int]) -> set:
    shingles = []
    q = len(s)
    for n in klist:
        if n > 0:
            shingles.append(set([s[i:(i + n)] for i in range(q - n + 1)]))
    return set.union(*shingles)



from typing import List, Set
import itertools


def replace_with_wildcard(word: str, indices: List[int],
                          wildcard: str = '\uFFFF'):
    """Replace char at a certain index with a wildcard

    word : str
        A string

    indices : List[int]
        List of indicies. len(word)<max(indices)

    wildcard : str
        An unicode char that is not actually used by any natural language,
          or the text that you analyzing, e.g. U+FFFE or U+FFFF
          See https://en.wikipedia.org/wiki/Specials_(Unicode_block)

    Example:
    --------
        word = 'ABCDE'
        replace_with_wildcard(s, [2, 4])
    """
    return ''.join([wildcard if i in indices else c
                    for i, c in enumerate(word)])


def get_wildcard_variants(word: str,
                          n_max_wildcards: int = None,
                          wildcard: str = '\uFFFF') -> List[str]:
    """Generate all wildcard variants for one word

    word : str
        A string

    n_max_wildcards : int
        Maximum number of wildcard characters per word.

    wildcard : str
        An unicode char that is not actually used by any natural language,
          or the text that you analyzing, e.g. U+FFFE or U+FFFF
          See https://en.wikipedia.org/wiki/Specials_(Unicode_block)
    """
    # Set the maximum number of number of wildcards for our word
    if n_max_wildcards is None:
        n_wild = len(word) - 1
    else:
        n_wild = min(len(word) - 1, n_max_wildcards)

    # pre-specify the word's indicies from which we draw combinations
    availidx = list(range(len(word)))

    # start loop
    variants = []
    for w in range(1, n_wild + 1):
        for indices in list(itertools.combinations(availidx, w)):
            variants.append(replace_with_wildcard(
                word, indices=indices, wildcard=wildcard))
    # done
    return variants


def wildcard_shinglesets(words: List[str],
                         n_max_wildcards: int = None,
                         wildcard: str = '\uFFFF') -> Set[str]:
    """Generate wildcard variants for each word

    word : str
        A string

    n_max_wildcards : int
        Maximum number of wildcard characters per word.

    wildcard : str
        An unicode char that is not actually used by any natural language,
          or the text that you analyzing, e.g. U+FFFE or U+FFFF
          See https://en.wikipedia.org/wiki/Specials_(Unicode_block)

    Example:
    --------
        import kshingle as ks
        shingles = ks.shingleset_k("aBc DeF", k=5)
        shingles = shingles.union(
            ks.wildcard_shinglesets(shingles, n_max_wildcards=2))
    """
    tmp = []
    for word in words:
        tmp.extend(get_wildcard_variants(word, n_max_wildcards, wildcard))
    return set(tmp)
